# Root package, with some helpers
# SPDX-FileCopyrightText: 2025 Nguyễn Gia Phong
# SPDX-License-Identifier: AGPL-3.0-or-later

from argparse import HelpFormatter, ONE_OR_MORE
from functools import cache
from importlib.resources import files

__all__ = ['__version__', 'GNUHelpFormatter', 'NetLoc',
           'atom2xhtml', 'format_epilog', 'format_version']
__version__ = '0.2.2'

EXAMPLE_PREFIX = ' ' * 2
# help2man's implementation detail
EXAMPLE_DESCRIPTION_PREFIX = ' ' * 20


class GNUHelpFormatter(HelpFormatter):
    """Help formatter for ArgumentParser following GNU Coding Standards."""

    def add_usage(self, usage, actions, groups, prefix='Usage: '):
        """Substitute 'Usage:' for 'usage:'."""
        super().add_usage(usage, actions, groups, prefix)

    def start_section(self, heading):
        """Substitute 'Options:' for 'options:'."""
        super().start_section(heading.capitalize())

    def _fill_text(self, text, width, indent):
        """Preserve examples' formatting."""
        desc_position = max(len(EXAMPLE_DESCRIPTION_PREFIX),
                            min(self._action_max_length+2,
                                self._max_help_position))
        desc_width = width - desc_position
        desc_indent = indent + ' '*desc_position
        example_indent = indent + EXAMPLE_PREFIX
        parts = []
        for line in text.splitlines():
            if line.startswith(EXAMPLE_DESCRIPTION_PREFIX):
                parts.append(super()._fill_text(line, desc_width, desc_indent))
            elif line.startswith(EXAMPLE_PREFIX):
                parts.append(example_indent+line.strip())
            else:  # not example
                parts.append(super()._fill_text(line, width, indent))
        return '\n'.join(parts)

    def _format_args(self, action, default_metavar):
        """Substitute 'METAVAR...' for 'METAVAR [METAVAR ...]'."""
        if action.nargs == ONE_OR_MORE:
            get_metavar = self._metavar_formatter(action, default_metavar)
            return '{}...'.format(*get_metavar(1))
        return super()._format_args(action, default_metavar)

    def _format_action_invocation(self, action):
        """Format --long-option=argument."""
        if not action.option_strings or action.nargs is not None:
            return super()._format_action_invocation(action)
        arg = self._format_args(action,
                                self._get_default_metavar_for_optional(action))
        return ', '.join(f"{opt}{'=' if opt.startswith('--') else ' '}{arg}"
                         for opt in action.option_strings)

    def add_argument(self, action):
        """Suppress positional arguments."""
        if action.option_strings:
            super().add_argument(action)


class NetLoc:
    """Network location with default port."""

    def __init__(self, default_port):
        self.default_port = default_port

    def __call__(self, string):
        """Return hostname and port from given netloc."""
        if ':' not in string:
            return string, self.default_port
        hostname, port = string.rsplit(':', 1)
        return hostname, int(port)  # ValueError to be handled by argparse


@cache
def atom2xhtml():
    """Load stylesheet from package resources exactly once."""
    return files(__name__).joinpath('atom2xhtml.xslt').read_bytes()


def format_epilog(examples):
    """Format example commands and their description ."""
    lines = ['Examples:']
    for example, description in examples:
        lines.append(EXAMPLE_PREFIX+example)
        lines.append(EXAMPLE_DESCRIPTION_PREFIX+description)
    lines.append('\nReport bugs to <chung@loa.loang.net>.')
    return '\n'.join(lines)


def format_version(prog, copyright_years, authors):
    """Format version information."""
    return (f'{prog} {__version__}\n\n'
            f'Copyright (C) {copyright_years} {authors}.\n\n'
            'This is free software: you are can redistribute and/or modify it'
            ' under the terms of the GNU Affero General Public License'
            ' version 3 or later <https://gnu.org/licenses/agpl>.\n\n'
            'There is NO WARRANTY, to the extent permitted by law.\n\n'
            f'Written by {authors}.')



# TLS certificate expiration checker
# SPDX-FileCopyrightText: 2025 Nguyễn Gia Phong
# SPDX-License-Identifier: AGPL-3.0-or-later

from argparse import ArgumentParser, FileType
from base64 import urlsafe_b64encode as base64
from datetime import datetime, timedelta, timezone
from email.utils import parsedate_to_datetime as parsedate
from itertools import chain
from os.path import basename
from socket import AF_INET, socket
from ssl import create_default_context as tls_context
from sys import argv, stderr, stdout
from unicodedata import category as unicode_category
from uuid import uuid4

from . import GNUHelpFormatter, NetLoc, format_epilog, format_version

__all__ = ['main']


def is_control_character(character):
    """Check if a Unicode character belongs to the control category."""
    return unicode_category(character) == 'Cc'


def printable(string):
    """Check if the given Unicode string is printable."""
    return not any(map(is_control_character, string))


def base64_from_str(string):
    """Convert string to base64 format in bytes."""
    return base64(string.encode()).decode()


def check(netlocs, after, output, fake_ca=None):
    """Check if each netloc's TLS certificate expires after given time.

    Print the certificate's summary to output file if that is the case.
    """
    ctx = tls_context()
    if fake_ca is not None:  # for testing
        fake_ca.configure_trust(ctx)

    for hostname, port in netlocs:
        now = datetime.now(timezone.utc).isoformat()
        netloc = f'{hostname}:{port}'
        stderr.write(f'TLS certificate for {netloc} ')
        try:
            with ctx.wrap_socket(socket(AF_INET),
                                 server_hostname=hostname) as conn:
                conn.connect((hostname, port))
                cert = conn.getpeercert()
        except Exception as exception:
            stderr.write(f'cannot be retrieved: {exception}\n')
            print(now, 'N/A', hostname, port, uuid4().int,
                  base64_from_str(str(exception)), file=output)
            continue

        try:
            not_before = parsedate(cert['notBefore'])
            not_after = parsedate(cert['notAfter'])
            ca = dict(chain.from_iterable(cert['issuer']))['organizationName']
            if not printable(ca):
                raise ValueError(f'CA name contains control character: {ca!r}')
            serial = int(cert['serialNumber'], 16)
        except Exception as exception:
            stderr.write(f'cannot be parsed: {exception}\n')
            print(now, 'N/A', hostname, port, uuid4().int,
                  base64_from_str(str(exception)), file=output)
        else:
            if after < not_after:
                after_seconds = after.isoformat(timespec='seconds')
                stderr.write(f'will not expire at {after_seconds}\n')
            else:
                stderr.write(f'will expire at {not_after.isoformat()}\n')
                print(not_before.isoformat(), not_after.isoformat(),
                      # As unique identifier
                      hostname, port, serial,
                      base64_from_str(ca), file=output)


def main(prog=basename(argv[0]), arguments=argv[1:]):
    """Run TLS checker."""
    desc = ('Check TLS certificate expiration of HOST,'
            ' where PORT defaults to 443.\n\n'
            'The output is intended to be used by scadere-listen(1).')
    examples = [((f'{prog} --output=/var/lib/scadere/certificates'
                  ' example.com example.net'),
                 ('check if TLS certificates used by example.com:443'
                  ' and example.net:443 are either invalid'
                  ' or expiring within the next week,'
                  ' then write the result to /var/lib/scadere/certificates.'))]

    parser = ArgumentParser(prog=prog, allow_abbrev=False, description=desc,
                            epilog=format_epilog(examples),
                            formatter_class=GNUHelpFormatter)
    parser.add_argument('-v', '--version', action='version',
                        version=format_version('scadere-check',
                                               '2025', 'Nguyễn Gia Phong'))
    parser.add_argument('netloc', metavar='HOST[:PORT]',
                        nargs='+', type=NetLoc(443))
    parser.add_argument('-d', '--days', type=float, default=7,
                        help='days before expiration (default to 7)')
    parser.add_argument('-o', '--output', metavar='PATH',
                        type=FileType('w'), default=stdout,
                        help='output file (default to stdout)')

    args = parser.parse_args(arguments)
    with args.output:  # pragma: no cover
        after = datetime.now(timezone.utc) + timedelta(days=args.days)
        check(args.netloc, after, args.output)


if __name__ == '__main__':  # pragma: no cover
    main()
